/**********************************************************************

seidel.h
Mark Cederholm (plp@pierssen.com)
Version 1.0

Class Seidel is an encapsulation of Seidel's polygon triangulation
algorithm.  It is adapted from the C code by Atul Narkhede and Dinesh
Manocha, which in turn updated Raimund Seidel's algorithm by taking
holes into account.  More information on the algorithm, including the
original C code, may be found at:

http://www.cs.unc.edu/~dm/CODE/GEM/chapter.html

The code has been improved by dynamically allocating memory for data
structures.  To deal with very large polygons you may need to increase
the call stack size when you compile the code.

-----------------------------------------------------------------------

Method triangulate_polygon has the following arguments:

   ncontours:  #contours

   cntr:  An array describing the number of points in each contour.  Thus,
   cntr[i] = #points in the i'th contour.

   vertices:  Input array of vertices.  Vertices for each contour
   immediately follow those for previous one.  Array location vertices[0]
   must NOT be used (i.e. i/p starts from vertices[1] instead.  The output
   triangles are specified w.r.t. the indices of these vertices.

   triangles:  Output array to hold triangles.

The method returns the number of triangles generated or 0 if failed.
Make sure an appropriate amount of memory is allocated for the triangle
array.

Every contour is specified by giving all its points in order.  No point
should be repeated. i.e. if the outer contour is a square, only the four
distinct endpoints should be specified in order.  Outer contour must be
anti-clockwise.  All inner contours must be clockwise.


Method is_point_inside_the_polygon may only be called after method
triangulate_polygon has been called.  It returns TRUE or FALSE depending
upon whether the vertex is inside the polygon or not.  This routine will
always detect all the points belonging to the set (polygon-area -
polygon-boundary).  The return value for points on the boundary is not
consistent!!!


Method reset clears all data and frees up memory.


**********************************************************************/


#include <malloc.h>
#include <math.h>
#include <memory.h>
#include <stdlib.h>
#include <time.h>


/**** Structure Types ****/


typedef struct {
  double x, y;
} point_t, vector_t;


/* Segment attributes */

typedef struct {
  point_t v0, v1;               /* two endpoints */
  int is_inserted;              /* inserted in trapezoidation yet ? */
  int root0, root1;             /* root nodes in Q */
  int next;                     /* Next logical segment */
  int prev;                     /* Previous segment */
} segment_t;


/* Trapezoid attributes */

typedef struct {
  int lseg, rseg;               /* two adjoining segments */
  point_t hi, lo;               /* max/min y-values */
  int u0, u1;
  int d0, d1;
  int sink;                     /* pointer to corresponding in Q */
  int usave, uside;             /* I forgot what this means */
  int state;
} trap_t;


/* Node attributes for every node in the query structure */

typedef struct {
  int nodetype;                 /* Y-node or S-node */
  int segnum;
  point_t yval;
  int trnum;
  int parent;                   /* doubly linked DAG */
  int left, right;              /* children */
} node_t;


typedef struct {
  int vnum;
  int next;                     /* Circularly linked list  */
  int prev;                     /* describing the monotone */
  int marked;                   /* polygon */
} monchain_t;


typedef struct {
  point_t pt;
  int vnext[4];                 /* next vertices for the 4 chains */
  int vpos[4];                  /* position of v in the 4 chains */
  int nextfree;
} vertexchain_t;


/**** Class Declaration ****/


class Seidel
{
   public:

   Seidel();                    // constructor
   int triangulate_polygon(int ncontours, int * cntr,
      double (* vertices)[2], int (* triangles)[3]);
   int is_point_inside_polygon(double vertex[2]);
   void reset();                // clear all data and free memory

   private:

   // data

   segment_t * seg;                   // Segment table
   node_t * qs;                       // Query structure
   trap_t * tr;                       // Trapezoid structure
   monchain_t * mchain;               // Monotone polygon table
   vertexchain_t * vert;              // chain init. information
   int * mon;                         // vertex position array
   int * visited;
   int * permute;                     // random segment array
   int SEGSIZE;
   int QSIZE;
   int TRSIZE;
   int q_idx, tr_idx;
   int chain_idx, op_idx, mon_idx;
   int choose_idx;
   int ntriangles;

   // methods

   int initialise(int);

   // construct.c

   int newnode();
   int newtrap();
   int _max(point_t *, point_t *, point_t *);
   int _min(point_t *, point_t *, point_t *);
   int _greater_than(point_t *, point_t *);
   int _equal_to(point_t *, point_t *);
   int _greater_than_equal_to(point_t *, point_t *);
   int _less_than(point_t *, point_t *);
   int init_query_structure(int);
   int is_left_of(int, point_t *);
   int inserted(int, int);
   int locate_endpoint(point_t *, point_t *, int);
   int merge_trapezoids(int, int, int, int);
   int add_segment(int);
   int find_new_roots(int);
   int construct_trapezoids(int);

   // monotone.c

   int inside_polygon(trap_t *);
   int newmon();
   int new_chain_element();
   double get_angle(point_t *, point_t *, point_t *);
   int get_vertex_positions(int, int, int *, int *);
   int make_new_monotone_poly(int, int, int);
   int monotonate_trapezoids(int);
   int traverse_polygon(int, int, int, int);
   int triangulate_monotone_polygons(int, int, int (*)[3]);
   int triangulate_single_polygon(int, int, int, int (*)[3]);

   // misc.c

   int generate_random_ordering(int);
   int choose_segment(void);
   int math_logstar_n(int);
   int math_N(int, int);

};


/**** Constants ****/


/* Node types */

#define T_X     1
#define T_Y     2
#define T_SINK  3


#define TRUE  1
#define FALSE 0

#define FIRSTPT 1               /* checking whether pt. is inserted */
#define LASTPT  2

#define INFINITY 1<<30
#define C_EPS 1.0e-7            /* tolerance value: Used for making */
                                /* all decisions about collinearity or */
                                /* left/right of segment. Decrease */
                                /* this value if the input points are */
                                /* spaced very close together */

#define S_LEFT 1                /* for merge-direction */
#define S_RIGHT 2

#define ST_VALID 1              /* for trapezium state */
#define ST_INVALID 2

#define SP_SIMPLE_LRUP 1        /* for splitting trapezoids */
#define SP_SIMPLE_LRDN 2
#define SP_2UP_2DN     3
#define SP_2UP_LEFT    4
#define SP_2UP_RIGHT   5
#define SP_2DN_LEFT    6
#define SP_2DN_RIGHT   7
#define SP_NOSPLIT    -1

#define TR_FROM_UP 1            /* for traverse-direction */
#define TR_FROM_DN 2

#define TRI_LHS 1
#define TRI_RHS 2


/**** Functions ****/


#define MAX(a, b) (((a) > (b)) ? (a) : (b))

#define MIN(a, b) (((a) < (b)) ? (a) : (b))

#define CROSS(v0, v1, v2) (((v1).x - (v0).x)*((v2).y - (v0).y) - \
                           ((v1).y - (v0).y)*((v2).x - (v0).x))

#define DOT(v0, v1) ((v0).x * (v1).x + (v0).y * (v1).y)

#define FP_EQUAL(s, t) (fabs(s - t) <= C_EPS)

#define CROSS_SINE(v0, v1) ((v0).x * (v1).y - (v1).x * (v0).y)

#define LENGTH(v0) (sqrt((v0).x * (v0).x + (v0).y * (v0).y))

#define log2(v0) (log(v0) / log(2))

