// triang.cpp

#include "triang.h"

int compare(const void * elem1, const void * elem2);

int Triang::do_it(char * a_shp_in, char * a_shp_out)
{
   bool result;

   shp_in = a_shp_in;
   shp_out = a_shp_out;

   ncontours = read_segments();
   if (ncontours == 0)
      return 1; // ERROR reading shapefile

   if (has_redundant_vertices())
   {
      free(cntr);
      free(vertices);
      return 2; // polygon touches self
   }

   ntriangles = gen_triangles();
   free(cntr);
   free(vertices);
   if (ntriangles == 0)
      return 3; // ERROR in triangulation

   result = write_triangles();
   free(triangles);
   if (! result)
      return 4; // ERROR writing shapefile

   return 0;
}


/**********************************************************************

This routine generates the master contour and vertex table from the
first polygon record of the input shapefile.

**********************************************************************/


int Triang::read_segments()
{

   int NumParts;
   int NumPoints;

   FILE * infile;
   int ibuf[1];
   fpos_t pos;
   size_t numread;
   int * pbuf;
   double * vbuf;

   int i,j,n;
   double x,y;
   int i0,i1,j0,j1;

   // Open input file.

   if ((infile = fopen(shp_in,"rb")) == NULL)
      return 0;

   // Skip file header and record header.

   pos = 108;
   if (fsetpos(infile,&pos) != 0)
      return 0;

   // Check for correct shapetype.

   numread = fread(ibuf,sizeof(int),1,infile);
   if (ibuf[0] != 5)
      return 0;

   // Get bounding box (use this later when writing triangles)

   numread = fread(bounds,sizeof(double),4,infile);
   if (numread != 4)
      return 0;

   // Get number of parts

   pos = 144;
   if (fsetpos(infile,&pos) != 0)
      return 0;
   numread = fread(ibuf,sizeof(int),1,infile);
   if (numread != 1)
      return 0;
   NumParts = ibuf[0];

   // Get number of points

   numread = fread(ibuf,sizeof(int),1,infile);
   if (numread != 1)
      return 0;
   NumPoints = ibuf[0];

   // Allocate space for parts buffer and read values

   pbuf = (int *) calloc(NumParts, sizeof(int));
   if (pbuf == NULL)
      return 0;
   numread = fread(pbuf,sizeof(int),NumParts,infile);
   if (numread != (size_t) NumParts)
      return 0;

   // Allocate space for vertex buffer, read values, and close infile

   vbuf = (double *) calloc(NumPoints, 2 * sizeof(double));
   if (vbuf == NULL)
      return 0;
   numread = fread(vbuf,sizeof(double),2 * NumPoints,infile);
   if (numread != (size_t) (NumPoints * 2))
      return 0;
   fclose(infile);

   // Allocate space for contour and vertex table

   cntr = (int *) calloc(NumParts, sizeof(int));
   if (cntr == NULL)
      return 0;
   nvertices = NumPoints - NumParts;
   vertices = (double (*)[2]) calloc(nvertices, 2 * sizeof(double));
   if (vertices == NULL)
      return 0;

   /*******************************************************************

   Copy the coordinates into the vertex list, reversing the order of each
   part, stripping the redundant vertex, and building the contour table entry
   for each

   *******************************************************************/

   j0 = 0;
   for (i = 0; i < NumParts; i++)
   {
      i0 = pbuf[i];
      if (i == (NumParts - 1))
         n = NumPoints - i0 - 1;
      else
         n = pbuf[i + 1] - i0 - 1;
      for (j = 0; j < n; j++)
      {
         i1 = (i0 + j) * 2;
         x = vbuf[i1];
         y = vbuf[i1 + 1];
         j1 = j0 + n - j - 1;
         vertices[j1][0] = x;
         vertices[j1][1] = y;
      }
      j0 += n;
      cntr[i] = n;
   }

   free(pbuf);
   free(vbuf);

   return NumParts;

}


/**********************************************************************

Because the Seidel triangulation blows up if a polygon has any redundant
vertices, this routine traps that.

**********************************************************************/


bool Triang::has_redundant_vertices()
{
   double (* vt)[2];
   bool redundant = false;
   int i;
   double x0, y0, dx, dy;

   // Allocate memory for a sortable vertex list

   vt = (double (*)[2]) calloc(nvertices, 2 * sizeof(double));
   if (vt == NULL)
      return false;

   // Populate the vertex list

   for (i = 0; i < nvertices; i++)
   {
      vt[i][0] = vertices[i][0];
      vt[i][1] = vertices[i][1];
   }

   // Sort it

   qsort(vt, nvertices, 2 * sizeof(double), compare);

   // Check for redundant vertices

   x0 = vt[0][0];
   y0 = vt[0][1];
   for (i = 1; i < nvertices; i++)
   {
      dx = vt[i][0] - x0;
      dy = vt[i][1] - y0;
      if ((dx < C_EPS) && (dy < C_EPS))
      {
         redundant = true;
         break;
      }
      x0 = vt[i][0];
      y0 = vt[i][1];
   }

   free(vt);
   return redundant;
}



int Triang::gen_triangles()
{

   int i, j, j0, r0;
   int numtri;
   int n, nc, nt, SEGSIZE;
   int numrings = 0;
   double vertex[2];
   Seidel tri;
   int * rings;
   double area;

   /*******************************************************************

   Because a shapefile polygon may have multiple contours, not all of which
   are islands, we need to classify them.  First, set up ring and island
   arrays and count rings with negative vs. positive areas.

   *******************************************************************/

   islands = (int *) calloc(ncontours, sizeof(int));
   if (islands == NULL)
      return 0;
   rings = (int *) calloc(ncontours, sizeof(int));
   if (rings == NULL)
      return 0;
   j0 = 0;
   r0 = 0;
   for (i = 0; i < ncontours; i++)
   {
      n = cntr[i];
      area = get_ring_area(j0,n);
      if (area < 0)
         islands[i] = 0;
      else
      {
         islands[i] = -1;
         rings[r0] = i;
         r0++;
         numrings++;
      }
      j0 += n;

   }

   // Allocate memory for temporary cntr array.

   c = (int *) calloc(ncontours, sizeof(int));
   if (c == NULL)
      return 0;

   // Allocate memory for temporary vertex and triangle arrays.
   // Add 256 to SEGSIZE just to play it safe.

   SEGSIZE = nvertices + 256;
   t = (int (*)[3]) calloc(SEGSIZE, 3 * sizeof(int));
   if (t == NULL)
      return 0;
   v = (double (*)[2]) calloc(SEGSIZE, 2 * sizeof(double));
   if (v == NULL)
      return 0;

   /*******************************************************************

   If only one outer ring exists, just do a single triangulation run.

   *******************************************************************/

   if (numrings == 1)
   {

      // Build vertex array

      for (i = 0; i < nvertices; i++)
      {
         v[i + 1][0] = vertices[i][0];
         v[i + 1][1] = vertices[i][1];

      }

      // Perform the triangulation

      numtri = tri.triangulate_polygon(ncontours, cntr, v, t);
      tri.reset();
      if (numtri == 0)
         return 0;

      // Build triangle coordinate array

      triangles = (double (*)[6]) calloc(numtri, 6 * sizeof(double));
      if (triangles == NULL)
         return 0;
      for (i = 0; i < numtri; i++)
      {
         triangles[i][0] = v[t[i][0]][0];
         triangles[i][1] = v[t[i][0]][1];
         triangles[i][2] = v[t[i][1]][0];
         triangles[i][3] = v[t[i][1]][1];
         triangles[i][4] = v[t[i][2]][0];
         triangles[i][5] = v[t[i][2]][1];

      }
      free(v);
      free(t);
      free(c);
      free(rings);
      free(islands);
      return numtri;

   }

   /*******************************************************************

   The remaining situation is trickier.  What we need to do is find out
   which islands belong to which outer ring.  That means triangulating each
   outer ring and doing a point-in-poly analysis of the islands.

   *******************************************************************/

   for (i = 0; i < numrings; i++)
   {
      r0 = rings[i];
      nc = build_ring_array(r0, false);
      numtri = tri.triangulate_polygon(nc, c, v, t);
      j0 = 0;
      for (j = 0; j < ncontours; j++)
      {
         n = cntr[i];
         if (islands[j] == 0)
         {
            vertex[0] = vertices[j0][0];
            vertex[1] = vertices[j0][1];
            if (tri.is_point_inside_polygon(vertex))
               islands[j] = r0;
         }
         j0 += n;
      }
      tri.reset();
   }

   /*******************************************************************

   Now that we have established the spatial relationship of the various
   parts, we need to build arguments for the triangulation of each.  The
   resulting triangles are merged into the master list.

   *******************************************************************/

   j0 = 0;
   for (i = 0; i < numrings; i++)
   {
      r0 = rings[i];
      nc = build_ring_array(r0, true);
      nt = tri.triangulate_polygon(nc, c, v, t);
      tri.reset();
      if (j0 == 0)
      {

         // Initialize triangle array

         triangles = (double (*)[6]) malloc(nt * 6 * sizeof(double));
         if (triangles == NULL)
            return 0;

      }
      else
      {

         // Expand triangle array

         n = _msize(triangles);
         triangles = (double (*)[6]) realloc(triangles, n + (nt * 6 * sizeof(double)));
         if (triangles == NULL)
            return 0;

      }

      // Add triangles

      for (j = 0; j < nt; j++)
      {
         triangles[j0 + j][0] = v[t[j][0]][0];
         triangles[j0 + j][1] = v[t[j][0]][1];
         triangles[j0 + j][2] = v[t[j][1]][0];
         triangles[j0 + j][3] = v[t[j][1]][1];
         triangles[j0 + j][4] = v[t[j][2]][0];
         triangles[j0 + j][5] = v[t[j][2]][1];

      }
      j0 += nt;

   }
   numtri = j0;
   free(v);
   free(t);
   free(c);
   free(rings);
   free(islands);
   return numtri;

}


/**********************************************************************

This routine writes a shapefile out of the resulting triangles.  A
single attribute called ID is created and populated with 0, much as does
ArcView when creating a new theme.

**********************************************************************/


bool Triang::write_triangles()
{

   char * idbuf = "           0";
   char path[_MAX_PATH];
   double bndbuf[4], t[8];
   FILE * shp, * shx, * dbf;
   int epos, i;
   size_t f_write, shpsize, shxsize;
   unsigned char bbuf[1];
   unsigned long ibuf[1], ibufs[1];

   // Open output files

   strcpy(path,shp_out);
   if ((shp = fopen(shp_out,"wb")) == NULL)
      return false;
   strcpy(path,shp_out);
   epos = strrchr(path, '.') - path + 1;
   strncpy(path + epos, "shx", 3);
   if ((shx = fopen(path,"wb")) == NULL)
   {
      _fcloseall();
      return false;
   }
   epos = strrchr(path, '.') - path + 1;
   strncpy(path + epos, "dbf", 3);
   if ((dbf = fopen(path,"wb")) == NULL)
   {
      _fcloseall();
      return false;
   }

   // Calculate sh file sizes

   shpsize = (ntriangles * 60) + 50;
   shxsize = (ntriangles * 4) + 50;

   // Write file headers

   write_sh_header(shp, shpsize);
   write_sh_header(shx, shxsize);
   write_db_header(dbf, ntriangles);

   // Write records

   for (i = 0; i < ntriangles; i++)
   {

      // .shp record number

      ibuf[0] = i + 1;
      ibufs[0] = endswap(ibuf[0]);
      f_write = fwrite(ibufs, sizeof(unsigned long), 1, shp);

      // record size

      ibuf[0] = 56;
      ibufs[0] = endswap(ibuf[0]);
      f_write = fwrite(ibufs, sizeof(unsigned long), 1, shp);

      // shape type

      ibuf[0] = 5;
      f_write = fwrite(ibuf, sizeof(unsigned long), 1, shp);

      // bounding box -- no need for anything elegant here

      bndbuf[0] = triangles[i][0];
      bndbuf[1] = triangles[i][1];
      bndbuf[2] = triangles[i][0];
      bndbuf[3] = triangles[i][1];

      if (triangles[i][2] < bndbuf[0])
         bndbuf[0] = triangles[i][2];
      if (triangles[i][2] > bndbuf[2])
         bndbuf[2] = triangles[i][2];
      if (triangles[i][3] < bndbuf[1])
         bndbuf[1] = triangles[i][3];
      if (triangles[i][3] > bndbuf[3])
         bndbuf[3] = triangles[i][3];
      if (triangles[i][4] < bndbuf[0])
         bndbuf[0] = triangles[i][4];
      if (triangles[i][4] > bndbuf[2])
         bndbuf[2] = triangles[i][4];
      if (triangles[i][5] < bndbuf[1])
         bndbuf[1] = triangles[i][5];
      if (triangles[i][5] > bndbuf[3])
         bndbuf[3] = triangles[i][5];

      f_write = fwrite(bndbuf, sizeof(double), 4, shp);

      // NumParts

      ibuf[0] = 1;
      f_write = fwrite(ibuf, sizeof(unsigned long), 1, shp);

      // NumPoints

      ibuf[0] = 4;
      f_write = fwrite(ibuf, sizeof(unsigned long), 1, shp);

      // Parts

      ibuf[0] = 0;
      f_write = fwrite(ibuf, sizeof(unsigned long), 1, shp);

      // Points (reverse order with redundant vertex)
      // Again, no need for anything elegent, just do it

      t[0] = triangles[i][4];
      t[1] = triangles[i][5];
      t[2] = triangles[i][2];
      t[3] = triangles[i][3];
      t[4] = triangles[i][0];
      t[5] = triangles[i][1];
      t[6] = triangles[i][4];
      t[7] = triangles[i][5];
      f_write = fwrite(t, sizeof(double), 8, shp);

      // .shx offset

      ibuf[0] = 50 + (60 * i);
      ibufs[0] = endswap(ibuf[0]);
      f_write = fwrite(ibufs, sizeof(unsigned long), 1, shx);

      // record size

      ibuf[0] = 56;
      ibufs[0] = endswap(ibuf[0]);
      f_write = fwrite(ibufs, sizeof(unsigned long), 1, shx);

      // .dbf record

      f_write = fwrite(idbuf, sizeof(char), 12, dbf);

   }

   // .dbf terminating byte

   bbuf[0] = 26;
   f_write = fwrite(bbuf, sizeof(unsigned char), 1, dbf);

   _fcloseall();
   return true;

}


double Triang::get_ring_area(int i0, int n)
{
   int i;
   double area = 0;
   double a0, a1, b0, b1, c0, c1;


   a0 = vertices[i0][0];
   a1 = vertices[i0][1];
   for (i = 1; i < (n - 1); i++)
   {
      b0 = vertices[i0 + i][0];
      b1 = vertices[i0 + i][1];
      c0 = vertices[i0 + i + 1][0];
      c1 = vertices[i0 + i + 1][1];
      area += (a0 * b1) - (a1 * b0) + (a1 * c0) - (a0 * c1) + (b0 * c1) - (c0 * b1);
   }
   return area;
}


int Triang::build_ring_array(int ring_id, bool add_islands)
{

   int i, j, n, i0, j0;
   int nc;

   i0 = 1;  // leave the first point blank
   j0 = 0;
   nc = 0;
   for (i = 0; i < ncontours; i++)
   {
      n = cntr[i];
      if ((i == ring_id) || (add_islands && (islands[i] == ring_id)))
      {
         for (j = 0; j < n; j++)
         {
            v[i0 + j][0] = vertices[j0 + j][0];
            v[i0 + j][1] = vertices[j0 + j][1];
         }
         i0 += n;
         c[nc] = n;
         nc++;
      }
      j0 += n;

   }

   /* debug
   if (add_islands)
   {
      j0 = 1;
      for (i = 0; i < nc; i++)
      {
         n = c[i];
         printf("part #%d: %d points\n",i,n);
         for (j = 0; j < n; j++)
            printf("%f@%f\n",v[j0 + j][0],v[j0 + j][1]);
         j0 += n;
      }
   }
   */

   return nc;

}


void Triang::write_sh_header(FILE * out_file, size_t fsize)
{

   size_t f_write;
   int i;
   unsigned long ibuf[1], ibufs[1];
   double dbuf[1];

   // File code

   ibuf[0] = 9994;
   ibufs[0] = endswap(ibuf[0]);
   f_write = fwrite(ibufs, sizeof(unsigned long), 1, out_file);

   // Unused space

   ibuf[0] = 0;
   for (i = 0; i < 5; i++)
      f_write = fwrite(ibuf, sizeof(unsigned long), 1, out_file);

   // File length

   ibuf[0] = fsize;
   ibufs[0] = endswap(ibuf[0]);
   f_write = fwrite(ibufs, sizeof(unsigned long), 1, out_file);

   // Version

   ibuf[0] = 1000;
   f_write = fwrite(ibuf, sizeof(unsigned long), 1, out_file);

   // Shape type

   ibuf[0] = 5;
   f_write = fwrite(ibuf, sizeof(unsigned long), 1, out_file);

   // Bounding box

   f_write = fwrite(bounds, sizeof(double), 4, out_file);
   dbuf[0] = 0;
   for (i = 0; i < 4; i++)
      f_write = fwrite(dbuf, sizeof(double), 1, out_file);

   return;

}


void Triang::write_db_header(FILE * out_file, int numrecs)
{

   int i;
   size_t f_write;
   struct tm * newtime;
   time_t long_time;
   unsigned char bbuf[1];
   unsigned long ibuf[1];
   unsigned short sbuf[1];

   // Initial byte

   bbuf[0] = 3;
   f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);

   // Creation date

   time(&long_time);
   newtime = localtime(&long_time);
   bbuf[0] = newtime->tm_year;
   f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);
   bbuf[0] = newtime->tm_mon + 1;
   f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);
   bbuf[0] = newtime->tm_mday;
   f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);

   // Number of records

   ibuf[0] = numrecs;
   f_write = fwrite(ibuf, sizeof(unsigned long), 1, out_file);

   // Number of bytes in header

   sbuf[0] = 65;
   f_write = fwrite(sbuf, sizeof(unsigned short), 1, out_file);

   // Number of bytes in record

   sbuf[0] = 12;
   f_write = fwrite(sbuf, sizeof(unsigned short), 1, out_file);

   // Remainder of header

   bbuf[0] = 0;
   for (i = 0; i < 17; i++)
      f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);
   bbuf[0] = 87;
   f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);
   bbuf[0] = 0;
   for (i = 0; i < 2; i++)
      f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);

   // Field descriptor

   bbuf[0] = 73;  // "I"
   f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);
   bbuf[0] = 68;  // "D"
   f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);
   bbuf[0] = 0;
   for (i = 0; i < 9; i++)
      f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);
   bbuf[0] = 78;  // "N"
   f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);
   bbuf[0] = 0;
   for (i = 0; i < 4; i++)
      f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);
   bbuf[0] = 11;  // width of field
   f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);
   bbuf[0] = 0;
   for (i = 0; i < 15; i++)
      f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);

   // Header terminator

   bbuf[0] = 13;
   f_write = fwrite(bbuf, sizeof(unsigned char), 1, out_file);

   return;

}


unsigned long Triang::endswap(unsigned long n)
{
   union
   {
      unsigned long p;
      unsigned char b[4];
   };
   unsigned char temp;
   int i;

   // Convert little-endian unsigned long to big-endian

   p = n;
   for (i = 0; i < 2; i++)
   {
      temp = b[i];
      b[i] = b[3-i];
      b[3-i] = temp;
   }

   return p;
}


int compare(const void * elem1, const void * elem2)
{

   double a;

   // vertex comparison routine for qsort

   a = (*(double(*)[2]) elem1)[0] - (*(double(*)[2]) elem2)[0];
   if (a > 0) return 1;
   if (a < 0) return -1;
   a = (*(double(*)[2]) elem1)[1] - (*(double(*)[2]) elem2)[1];
   if (a > 0) return 1;
   if (a < 0) return -1;
   return 0;

}

